#define DIMENSIONS 2
#define MATRIXTYPE CRSMatrix

#define GRAVITY_AXES 1  // 1=y, 0=x, -100 = NO GRAVITY
#define RHO 1.0
#define GRAVITY 1.0  // Please use g = 1.0 instead of 9.81 here.

#include <numeric>
#include <algorithm>
#include <ctime>

#include "CRSMatrix.hh"
#include "gw_assembler.hh"
#include "well_models.hh"
#include "particletracker.hh"


#define ELLIPTIC_VTK
#include "vtkwriter.hh"

using namespace std;
#define SOLVER bicgstabsor_solve


//
// Invoking the Groundwater Solver:
//
template <class PostProcessor, class SpecialBoundaryConditions>
Vector groundwater2d(
					 ConductivityCoefficients & conductivity,
					 SpecialBoundaryConditions & bc,
					 const Sources & sources,
					 const Grid & grid,
					 std::string output_filename
					 )
{
  cout << endl;
  cout << "=====================================================================" << endl;
  cout << "Computing solution for groundwater problem for grid size "
       << grid.cells()[0] <<"x"<< grid.cells()[1] << endl;
  
  // setup assembler
  GroundwaterAssembler assembler(grid,conductivity,sources,bc);

  // setup matrix and rhs
  typedef GroundwaterAssembler::Matrix Matrix;
  std::vector<size_t> size = grid.cells();
  const size_t grid_size = std::accumulate(size.begin(),size.end(),1,std::multiplies<size_t>());

  Matrix A(grid_size,5);
  A.set_omega(1.9);

  //Matrix A(grid_size,grid_size,0);
  Vector b(grid_size,0.0);

  // setup clock
  clock_t timer_begin = clock();
  clock_t timer;
  cout << "Begin assembling" << endl;
  assembler.assemble(A,b);

  timer = clock();
  cout << "Time for assembling: " << (double)(timer-timer_begin)/CLOCKS_PER_SEC << endl;
  timer_begin = clock();

  // solve LSE
  Vector x(grid_size,0.);
  
  A.cgssor_solve(x,b);

  timer = clock();
  cout << "Time for solving: " << (double)(timer-timer_begin)/CLOCKS_PER_SEC << endl;
  timer_begin = clock();

  VTKWriter vtkwriter(grid);
  vtkwriter.write_groundwater(output_filename.c_str(),x,conductivity,bc);

  return x;
}




Vector particle_tracking2d(
			   const ParticleTracker::ConvectionField & advection,
			   const ParticleTracker::DispersionField & dispersion,
			   std::vector<ParticleTracker::Particle> & particles,
			   const Grid & grid,
			   const double dxfactor, 
			   double endtime,
			   std::string output_filename, 
			   double output_tau,
			   bool step_output = false
			   )
{
  cout << endl;
  cout << "=====================================================================" << endl;
  cout << "Computing solution for Particle Tracking problem for "
       << particles.size() << " particles." << endl;
 
  Vector h(Grid::dimensions,0);
  double h_min(std::numeric_limits<double>::max());
  for(int i=0;i<Grid::dimensions;++i)
	{
	  h[i] = grid.extent()[i] / ( (double)grid.cells()[i] );
	  if(h[i] < h_min) h_min = h[i];
	}

  // Some constant spatial step lenght "delta_x":
  const double dx = h_min * dxfactor;

  ParticleTracker tracker( grid, advection, dispersion, particles );

  // Number of cells per dimension
  const std::vector<size_t> cells(grid.cells());
  // Number of all cells
  const size_t grid_size = std::accumulate(
					   cells.begin(),
					   cells.end(),
					   1,
					   std::multiplies<size_t>()
					   );

  size_t step = 0;
  double output_t = 0;

  do{
    if( !tracker.step( dx, output_t ) )
	  {
		// Initialize container for the concentration field.
		Vector concentration_field( grid_size, 0. );

		// Fill concentration field with appropriate values after each step of particle movement.
		tracker.get_concentration_field( concentration_field );
				      
		// Output a snapshot of the current concentration field.
		cout << "Writing VTK output for t=" << output_t << endl;
		VTKWriter vtkwriter(grid);
		char buffer[128];
		sprintf( buffer, "%05.0lf",(double)step );
		std::string step_filename = output_filename;
		step_filename += buffer;
      
		vtkwriter.write_scalar(
				       step_filename.c_str()
				       , concentration_field
				       );
		
		
		output_t += output_tau;
		if( output_t > endtime )
		  output_t = endtime;

		if( step_output )
		  {
		    cout << endl;
		    cout << "------------" << endl;
		    cout << "Step nr " << step << " at t = " << output_t << endl;
		  }
		step++;	
	  }

  } while( output_t < endtime );

  // Output a snapshot of the final concentration field.
  VTKWriter vtkwriter(grid);
  std::string final_filename = output_filename + string("_final");
  Vector concentration_field( grid_size, 0. );
  tracker.get_concentration_field( concentration_field );

  vtkwriter.write_scalar(
			 final_filename.c_str()
			 , concentration_field 
			 );

  cout << "Final time = " << output_t << endl;
  
  return concentration_field;
}





int main(int argc, char* argv[])
{
  double endtime = 1728000;
  double dtout = 86400;

  const std::string groundwater_solution_filename("gw_solution.dat");

  std::vector<size_t> size(Grid::dimensions);
  size[0]=512; size[1]=256;
  Vector extent(Grid::dimensions);
  extent[0]=51.2; extent[1]=25.6;
  Grid grid(extent,size);

  // Define well 
  const double epsilon = 1e-10;
  Vector well_begin(2,extent[0]);
  well_begin[0]*=0.5;
  well_begin[1]=10.-epsilon;
  Vector well_end(2,extent[0]);
  well_end[0]*=0.5;
  well_end[0]+=extent[0]/(double)(size[0]);
  well_end[1]=extent[1]+epsilon;

  // Zero sources
  Sources sources;

  typedef SpatialParameters ConductivityCoefficients;
  ConductivityCoefficients conductivity( "field.cimg", grid );

  const double well_conductivity = 1.0;

  // RE-USING Model 3:
  WellPostProcessor post_processor( well_begin, well_end, well_conductivity );

  post_processor.process( grid, conductivity );
  
  BoundaryConditions pressure_bc( grid, "bc_gw.cfg" );

  Vector groundwater_solution(size[0]*size[1]);

  std::ifstream file;
  file.open( groundwater_solution_filename.c_str() );

  if( file )
	{
	  cout << "Reading convection field from file \""
		   <<groundwater_solution_filename
		   <<"\"" << endl;
	  for(size_t i=0; i<groundwater_solution.size(); ++i)
		file >> groundwater_solution[i];
	  file.close();
	}
  else
	{	  
	  groundwater_solution = 
		groundwater2d<WellPostProcessor,BoundaryConditions> ( conductivity
															  , pressure_bc
															  , sources
															  , grid
															  , "groundwater" 
															  );
	  
	  std::ofstream file;
	  file.open( groundwater_solution_filename.c_str() );
	  for(size_t i=0; i<groundwater_solution.size(); ++i)
		file << groundwater_solution[i] << " ";
	  file.close();
	}
  cout << "Initializing Convection and Dispersion fields" << endl;

  PressureVelocityConverter advection( 
									  groundwater_solution
									  , pressure_bc
									  , grid
									  , conductivity
									  , 0.33                 // porosity
									   );
  ConstantDispersion dispersion( 4e-7 );
  
  size_t n_particles = 100;
  if(argc > 1)
    n_particles = atoi(argv[1]);
  
  double dxfactor = 0.2;
  if(argc > 2)
    dxfactor = atof(argv[2]);
  
  
  // INSERT YOUR CODE HERE TO INITIALIZE THE LOCATIONS OF ALL PARTICLES
  
  const size_t n_side_particles = (size_t) sqrt((double)n_particles);
  n_particles = n_side_particles * n_side_particles;
  std::vector<ParticleTracker::Particle> particles(n_particles);
  const double X = 3.0;
  const double spacing = X / n_side_particles;
  Vector offset(Grid::dimensions);
  offset[0] = 3.0; 
  offset[1] = 6.0;
  
  cout << "Initializing particles" << endl;
  
  size_t p = 0;
  for(size_t y=0; y<n_side_particles; ++y)
	{
	  for(size_t x=0; x<n_side_particles; ++x)
		{
		  particles[p].position[0] = spacing * (double)x + offset[0]; 
		  particles[p].position[1] = spacing * (double)y + offset[1]; 
		  ++p;
		}
	}
  
  particle_tracking2d(
					  advection
					  , dispersion
					  , particles
					  , grid
					  , dxfactor
					  , endtime
					  , "transport"
					  , dtout
					  , true
					  );
  
  return 0;
}
